package fxGui.debugger.variables

import javafx.beans.property.ReadOnlyStringPropertyBase
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.launch
import java.lang.reflect.Field

/**
 * JavaFX string property using reflection to periodically check for value changes.
 *
 * @author Konstantin Kopper
 * @since 2.0.0
 * @param name See [getName].
 * @param bean See [getBean].
 * @constructor Creates a new property to watch a field.
 * @param f The field to be monitored.
 * @param instance The instance of which the field should be monitored.
 */
class ReflectionStringProperty(
    f: Field,
    instance: Class<*>,
    scope: CoroutineScope,
    private val name: String = f.name,
    private val bean: Any? = null
) : ReadOnlyStringPropertyBase(), ReflectionProperty<String>, CoroutineScope by scope {

    /**
     * Function which produces the current value of the underlying field.
     *
     * @author Konstantin Kopper
     * @since 2.0.0
     */
    private val getter: () -> String = { f.get(instance) as String }

    init {
        require(f.type == String::class.java) { "Given field does not have string type." }
        require(f in instance.fields) { "Given field not found in concrete instance." }

        // Coroutine can be started here. Cancellation ensured by implementing CoroutineScope.
        // For debugging: add CoroutineName("observe_string:${f.name}") as context parameter
        launch { super.observe(::fireValueChangedEvent) }
    }

    override fun get(): String = getter()
    override fun getName(): String = name
    override fun getBean(): Any? = bean
}
