package jani.interaction.basic

import kotlinx.serialization.KSerializer
import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.SerializationException
import kotlinx.serialization.descriptors.SerialDescriptor
import kotlinx.serialization.descriptors.buildClassSerialDescriptor
import kotlinx.serialization.encoding.Decoder
import kotlinx.serialization.encoding.Encoder
import kotlinx.serialization.json.JsonDecoder
import kotlinx.serialization.json.JsonNull
import kotlinx.serialization.json.JsonPrimitive
import kotlinx.serialization.json.booleanOrNull
import kotlinx.serialization.json.doubleOrNull
import kotlinx.serialization.json.intOrNull

@Serializable
data class ParameterDefinition(
    val id: Identifier,
    val name: String,
    val description: String?,
    val category: String?,
    @SerialName("is-global") val isGlobal: Boolean,
    val type: ParameterType,
    @SerialName("default-value") val defaultValue: List<DefaultValue>
) {
    init {
        // TODO do some checks
    }

    @Serializable(with = DefaultValue.Companion::class)
    sealed class DefaultValue {
        object Null : DefaultValue()
        object True : DefaultValue()
        object False : DefaultValue()
        sealed class Number : DefaultValue() {
            data class Integer(val n: Int) : Number()
            data class Real(val n: Double) : Number()
        }

        data class String(val s: kotlin.String) : DefaultValue()

        /**
         * Serializer for [DefaultValue].
         * Uses JSON inspection for type determination during deserialization.
         *
         * @author Konstantin Kopper
         * @see kotlinx.serialization.json.JsonContentPolymorphicSerializer
         */
        companion object : KSerializer<DefaultValue> {
            override val descriptor: SerialDescriptor =
                buildClassSerialDescriptor("jani.interaction.basic.ParameterDefinition.DefaultValue")

            override fun deserialize(decoder: Decoder): DefaultValue {
                val jsonDecoder =
                    decoder as? JsonDecoder ?: throw SerializationException("Only JSON can be deserialized.")
                return when (val element = jsonDecoder.decodeJsonElement()) {
                    is JsonNull -> Null
                    is JsonPrimitive -> element.booleanOrNull?.let { if (it) True else False }
                        ?: element.intOrNull?.let { Number.Integer(it) }
                        ?: element.doubleOrNull?.let { Number.Real(it) }
                        ?: String(element.content)
                    else -> throw SerializationException("Expected 'null' of JSON primitive.")
                }
            }

            @Suppress("EXPERIMENTAL_API_USAGE")
            override fun serialize(encoder: Encoder, value: DefaultValue) = when (value) {
                is Null -> encoder.encodeNull()
                is True -> encoder.encodeBoolean(true)
                is False -> encoder.encodeBoolean(false)
                is Number.Integer -> encoder.encodeInt(value.n)
                is Number.Real -> encoder.encodeDouble(value.n)
                is String -> encoder.encodeString(value.s)
            }
        }
    }
}
