package jani.interaction.basic

import kotlinx.serialization.KSerializer
import kotlinx.serialization.Required
import kotlinx.serialization.Serializable
import kotlinx.serialization.SerializationException
import kotlinx.serialization.builtins.ListSerializer
import kotlinx.serialization.builtins.serializer
import kotlinx.serialization.descriptors.SerialDescriptor
import kotlinx.serialization.descriptors.buildClassSerialDescriptor
import kotlinx.serialization.encoding.Decoder
import kotlinx.serialization.encoding.Encoder
import kotlinx.serialization.json.JsonArray
import kotlinx.serialization.json.JsonDecoder
import kotlinx.serialization.json.JsonObject
import kotlinx.serialization.json.JsonPrimitive
import kotlinx.serialization.json.jsonPrimitive

@Serializable(with = ParameterType.Companion::class)
sealed class ParameterType {

    object Bool : ParameterType()
    object Int : ParameterType()
    object Real : ParameterType()
    object String : ParameterType()

    @Serializable(with = Enum.Companion::class)
    class Enum(values: List<kotlin.String>) : ParameterType(), List<kotlin.String> by values {
        init {
            assert(values.isNotEmpty())
        }

        constructor(vararg values: kotlin.String) : this(values.asList())

        /**
         * Serializer for [Enum]s.
         * Uses default serializer for [List]s of [String]s.
         *
         * @author Konstantin Kopper
         * @see ListSerializer
         */
        internal companion object : KSerializer<Enum> {
            private val listSerializer = ListSerializer(kotlin.String.serializer())

            override val descriptor: SerialDescriptor = listSerializer.descriptor

            override fun deserialize(decoder: Decoder): Enum = Enum(decoder.decodeSerializableValue(listSerializer))

            override fun serialize(encoder: Encoder, value: Enum) =
                encoder.encodeSerializableValue(listSerializer, value)
        }
    }

    @Serializable
    data class Option(val type: ParameterType) : ParameterType() {
        @Required
        val kind = "option"

        init {
            assert(type !is Option)
        }
    }

    /**
     * Serializer for [ParameterType].
     * Uses JSON inspection for type determination during deserialization.
     *
     * @author Konstantin Kopper
     * @see kotlinx.serialization.json.JsonContentPolymorphicSerializer
     */
    companion object : KSerializer<ParameterType> {
        override val descriptor: SerialDescriptor = buildClassSerialDescriptor("jani.interaction.basic.ParameterType")

        override fun deserialize(decoder: Decoder): ParameterType {
            val jsonDecoder = decoder as? JsonDecoder ?: throw SerializationException("Only JSON can be deserialized.")
            return when (val element = jsonDecoder.decodeJsonElement()) {
                is JsonPrimitive -> stringTypes[element.jsonPrimitive.content]
                    ?: throw SerializationException("Unknown parameter type.")
                is JsonArray -> jsonDecoder.json.decodeFromJsonElement(Enum.serializer(), element)
                is JsonObject -> jsonDecoder.json.decodeFromJsonElement(Option.serializer(), element)
            }
        }

        override fun serialize(encoder: Encoder, value: ParameterType) =
            when (value) {
                is Bool, Int, Real, String -> encoder.encodeString(value::class.simpleName!!.lowercase())
                is Enum -> encoder.encodeSerializableValue(Enum.serializer(), value)
                is Option -> encoder.encodeSerializableValue(Option.serializer(), value)
            }

        private val stringTypes = mapOf("bool" to Bool, "int" to Int, "real" to Real, "string" to String)
    }
}
