package jani.interaction.basic.messages

import jani.interaction.tasks.ProvideTaskMessage
import jani.interaction.tasks.ProvideTaskProgress
import jani.interaction.tasks.ProvideTaskStatus
import jani.interaction.tasks.StopTask
import jani.interaction.tasks.TaskEnded
import jani.interaction.tasks.analyse.messages.ProvideAnalysisResults
import jani.interaction.tasks.analyse.messages.QueryAnalysisEngines
import jani.interaction.tasks.analyse.messages.ReplyAnalysisEngines
import jani.interaction.tasks.analyse.messages.StartAnalysisTask
import kotlinx.serialization.Serializable
import kotlinx.serialization.SerializationException
import kotlinx.serialization.modules.SerializersModule
import kotlinx.serialization.modules.polymorphic
import kotlinx.serialization.modules.subclass

@Serializable
abstract class JaniMessage {

    companion object {
        /**
         * [SerializersModule] for server-to-client messages according to the jani-interaction protocol.
         * Separation required as some query/reply pairs share the same type attribute,
         * which does not allow to choose the appropriate serializer automatically.
         *
         * @author Konstantin Kopper
         */
        val serverModule = SerializersModule {
            polymorphic(JaniMessage::class) {
                // Basic messages
                subclass(Capabilities::class)
                subclass(Close::class)
                subclass(ReplyUpdateServerParameters::class)
                // Tasks
                subclass(ProvideTaskStatus::class)
                subclass(ProvideTaskProgress::class)
                subclass(ProvideTaskMessage::class)
                subclass(TaskEnded::class)
                // Analyse tasks
                subclass(ReplyAnalysisEngines::class)
                subclass(ProvideAnalysisResults::class)
            }
        }

        /**
         * [SerializersModule] for client-to-server messages according to the jani-interaction protocol.
         * Separation required as some query/reply pairs share the same type attribute,
         * which does not allow to choose the appropriate serializer automatically.
         *
         * @author Konstantin Kopper
         */
        val clientModule = SerializersModule {
            polymorphic(JaniMessage::class) {
                // Basic messages
                subclass(Close::class)
                subclass(RequestUpdateServerParameters::class)
                // Tasks
                subclass(StopTask::class)
                // Analyse tasks
                subclass(QueryAnalysisEngines::class)
                subclass(StartAnalysisTask::class)
                default {
                    when (it) {
                        null -> Authenticate.serializer()
                        else -> throw SerializationException("Unsupported message type.")
                    }
                }
            }
        }
    }
}
