package jani.interaction.tasks.analyse

import jani.interaction.basic.Identifier
import kotlinx.serialization.KSerializer
import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.SerializationException
import kotlinx.serialization.descriptors.SerialDescriptor
import kotlinx.serialization.descriptors.buildClassSerialDescriptor
import kotlinx.serialization.encoding.Decoder
import kotlinx.serialization.encoding.Encoder
import kotlinx.serialization.json.JsonDecoder
import kotlinx.serialization.json.JsonObject
import kotlinx.serialization.json.JsonPrimitive
import kotlinx.serialization.json.booleanOrNull
import kotlinx.serialization.json.doubleOrNull
import kotlinx.serialization.json.intOrNull

@Serializable
data class AnalysisDataSet(
    val id: Int,
    val label: String?,
    val experiment: Identifier?,
    val property: Identifier?,
    val results: List<Result>?,
    @SerialName("data-sets") val dataSets: List<AnalysisDataSet>?

) {
    init {
        assert(id >= 1)
    }

    @Serializable
    data class Result(
        val label: String,
        val type: Type,
        val value: Value,
        val unit: Unit?,
        @SerialName("formatted-value") val formattedValue: String?
    ) {

        constructor(label: String, boolean: Boolean) :
            this(label, Type.BOOL, if (boolean) Value.True else Value.False, null, null)

        constructor(label: String, boolean: Boolean, formattedValue: String) :
            this(label, Type.BOOL, if (boolean) Value.True else Value.False, null, formattedValue)

        // TODO more constructors

        init {
            when (type) {
                Type.BOOL -> assert(value === Value.True || value === Value.False)
                Type.INT -> assert(value is Value.Number.Int)
                Type.RATIONAL -> assert(value is Value.Rational)
                Type.DECIMAL -> assert(value is Value.Number.Decimal)
                Type.STRING -> assert(value is Value.String)
            }
            when (unit) {
                Unit.SECOND, Unit.MILLISECOND -> assert(type === Type.INT || type === Type.RATIONAL || type === Type.DECIMAL)
                Unit.BYTE -> assert(type === Type.INT)
            }
        }

        @Serializable
        enum class Type {
            @SerialName("bool")
            BOOL,

            @SerialName("int")
            INT,

            @SerialName("rational")
            RATIONAL,

            @SerialName("decimal")
            DECIMAL,

            @SerialName("string")
            STRING
        }

        @Serializable(with = Value.Companion::class)
        sealed class Value {

            object True : Value()
            object False : Value()

            sealed class Number : Value() {
                data class Int(val n: kotlin.Int) : Number()
                data class Decimal(val d: Double) : Number()
            }

            @Serializable
            data class Rational(val num: Int, val den: Int) : Value() {
                init {
                    assert(den >= 0)
                }
            }

            data class String(val s: kotlin.String) : Value()

            companion object : KSerializer<Value> {
                override val descriptor: SerialDescriptor =
                    buildClassSerialDescriptor("jani.interaction.tasks.analyse.AnalysisDataSet.Value")

                override fun deserialize(decoder: Decoder): Value {
                    val jsonDecoder =
                        decoder as? JsonDecoder ?: throw SerializationException("Only JSON can be deserialized.")

                    return when (val element = jsonDecoder.decodeJsonElement()) {
                        is JsonPrimitive -> element.booleanOrNull?.let { if (it) True else False }
                            ?: element.intOrNull?.let { Number.Int(it) }
                            ?: element.doubleOrNull?.let { Number.Decimal(it) }
                            ?: String(element.content)
                        is JsonObject -> jsonDecoder.json.decodeFromJsonElement(Rational.serializer(), element)
                        else -> throw SerializationException("Unexpected JSON element.")
                    }
                }

                override fun serialize(encoder: Encoder, value: Value) = when (value) {
                    True -> encoder.encodeBoolean(true)
                    False -> encoder.encodeBoolean(false)
                    is Number.Int -> encoder.encodeInt(value.n)
                    is Number.Decimal -> encoder.encodeDouble(value.d)
                    is Rational -> encoder.encodeSerializableValue(Rational.serializer(), value)
                    is String -> encoder.encodeString(value.s)
                }
            }
        }

        @Serializable
        enum class Unit {
            @SerialName("s")
            SECOND,

            @SerialName("ms")
            MILLISECOND,

            @SerialName("B")
            BYTE
        }
    }
}
