package jani.interaction.tasks.analyse.messages

import jani.interaction.basic.AnyModel
import jani.interaction.basic.Identifier
import jani.interaction.basic.ParameterValue
import jani.interaction.tasks.JaniTask
import kotlinx.serialization.KSerializer
import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.SerializationException
import kotlinx.serialization.descriptors.SerialDescriptor
import kotlinx.serialization.descriptors.buildClassSerialDescriptor
import kotlinx.serialization.encoding.Decoder
import kotlinx.serialization.encoding.Encoder
import kotlinx.serialization.json.JsonDecoder
import kotlinx.serialization.json.JsonPrimitive
import kotlinx.serialization.json.booleanOrNull
import kotlinx.serialization.json.intOrNull

@Serializable
@SerialName("analysis-start")
class StartAnalysisTask(
    override val id: Int,
    val engine: Identifier,
    @SerialName("modelling-formalism") val modellingFormalism: Identifier?,
    val model: AnyModel,
    val properties: List<Identifier>?,
    val parameters: List<ParameterValue>,
    val experiments: List<Experiment>?
) : JaniTask() {

    constructor(id: Int, engine: Identifier, model: AnyModel, parameters: List<ParameterValue>) :
        this(id, engine, null, model, null, parameters, null)

    constructor(
        id: Int,
        engine: Identifier,
        modellingFormalism: Identifier?,
        model: AnyModel,
        parameters: List<ParameterValue>
    ) :
        this(id, engine, modellingFormalism, model, null, parameters, null)

    constructor(
        id: Int,
        engine: String,
        modellingFormalism: String?,
        model: AnyModel,
        parameters: List<ParameterValue>
    ) :
        this(
            id, Identifier(engine),
            if (modellingFormalism != null) Identifier(modellingFormalism) else null,
            model, parameters
        )

    init {
        assert(
            (model is AnyModel.JaniModel && modellingFormalism == null) ||
                (model is AnyModel.Textual && modellingFormalism != null)
        )
    }

    @Serializable
    data class Experiment(
        val experiment: Identifier,
        val values: List<Value>
    ) {
        @Serializable
        data class Value(
            val name: Identifier,
            val value: Value
        ) {
            @Serializable(with = Value.Companion::class)
            sealed class Value {
                object True : Value()
                object False : Value()
                data class Number(val n: Int) : Value()

                companion object : KSerializer<Value> {
                    override val descriptor: SerialDescriptor =
                        buildClassSerialDescriptor("jani.interaction.tasks.analyse.messages.StartAnalysisTask.Experiment.Value.Value")

                    override fun deserialize(decoder: Decoder): Value {
                        val jsonDecoder =
                            decoder as? JsonDecoder ?: throw SerializationException("Only JSON can be deserialized.")
                        val jsonPrimitive = jsonDecoder.decodeJsonElement() as? JsonPrimitive
                            ?: throw SerializationException("Expected a JSON primitive.")

                        return jsonPrimitive.booleanOrNull?.let { if (it) True else False }
                            ?: jsonPrimitive.intOrNull?.let { Number(it) }
                            ?: throw SerializationException("Unsupported JSON primitive.")
                    }

                    override fun serialize(encoder: Encoder, value: Value) = when (value) {
                        is True -> encoder.encodeBoolean(true)
                        is False -> encoder.encodeBoolean(false)
                        is Number -> encoder.encodeInt(value.n)
                    }
                }
            }
        }
    }
}
