package jani.model.models

import jani.model.Identifier
import kotlinx.serialization.KSerializer
import kotlinx.serialization.Serializable
import kotlinx.serialization.SerializationException
import kotlinx.serialization.descriptors.PrimitiveKind
import kotlinx.serialization.descriptors.PrimitiveSerialDescriptor
import kotlinx.serialization.encoding.Decoder
import kotlinx.serialization.encoding.Encoder

@Serializable(with = ModelFeature.ModelFeatureSerializer::class)
sealed class ModelFeature(val name: Identifier) {

    constructor(name: String) : this(Identifier(name))

    val asString
        get() = name.plainString

    object Arrays : ModelFeature("arrays")
    object Datatypes : ModelFeature("datatypes")
    object DerivedOperators : ModelFeature("derived-operators")
    object EdgePriorities : ModelFeature("edge-priorities")
    object Functions : ModelFeature("functions")
    object HyperbolicFunctions : ModelFeature("hyperbolic-functions")
    object NamedExpressions : ModelFeature("named-expressions")
    object NonDetSelection : ModelFeature("nondet-selection")
    object TradeoffProperties : ModelFeature("tradeoff-properties")
    object TrigonometricFunctions : ModelFeature("trigonometric-functions")

    override fun equals(other: Any?): Boolean = other != null && other is ModelFeature && name == other.name

    override fun hashCode(): Int = name.hashCode()

    private class ModelFeatureImpl(name: String) : ModelFeature(name) {
        init {
            require(name.startsWith("x-")) { "The name of custom model features must start with 'x-'." }
        }
    }

    companion object {
        fun new(name: String): ModelFeature = ModelFeatureImpl(name)
    }

    object ModelFeatureSerializer : KSerializer<ModelFeature> {
        override val descriptor = PrimitiveSerialDescriptor(ModelFeature::class.qualifiedName!!, PrimitiveKind.STRING)

        override fun deserialize(decoder: Decoder): ModelFeature = when (val str = decoder.decodeString()) {
            Arrays.asString -> Arrays
            Datatypes.asString -> Datatypes
            DerivedOperators.asString -> DerivedOperators
            EdgePriorities.asString -> EdgePriorities
            Functions.asString -> Functions
            HyperbolicFunctions.asString -> HyperbolicFunctions
            NamedExpressions.asString -> NamedExpressions
            NonDetSelection.asString -> NonDetSelection
            TradeoffProperties.asString -> TradeoffProperties
            TrigonometricFunctions.asString -> TrigonometricFunctions
            else -> try {
                new(str)
            } catch (e: IllegalArgumentException) {
                throw SerializationException("Creating a custom model feature failed.", e)
            }
        }

        override fun serialize(encoder: Encoder, value: ModelFeature) = encoder.encodeString(value.asString)
    }
}
