package pseuco.javaCompiler

import codeGen.Code
import codeGen.CodeGen
import codeGen.MainCode
import codeGen.MainCodeGen
import fxGui.Main
import org.apache.commons.io.FileUtils
import pseuco.PseuCoTask
import tree.ParseException
import tree.PseuCoParser
import typeChecker.TypeEnvironmentException
import typeChecker.TypeIllegalException
import util.FileUtilities
import util.ZipUtilities
import java.io.File
import java.io.OutputStream
import java.io.PrintStream
import java.io.PrintWriter
import javax.tools.DiagnosticCollector
import javax.tools.JavaFileObject
import javax.tools.ToolProvider

/**
 * Abstract task related to the pseuco-java-compiler.
 * Provides some common functionality, the rest needs to be implemented in concrete subclasses.
 *
 * @author Konstantin Kopper
 * @since 2.0.0
 * @see PseuCoTask
 * @property toBeExecuted The pseuCo file which should be executed by this task.
 * @property out The output stream to which the compiler output should be redirected.
 * @property err The error stream to which the compiler output should be redirected.
 * @constructor Create a new (abstract) task.
 */
abstract class PseuCoCompilerTask(
    toBeExecuted: File,
    protected val out: OutputStream,
    protected val err: OutputStream
) : PseuCoTask(toBeExecuted) {

    /**
     * Create a new (abstract) task using the standard output streams.
     *
     * @author Konstantin Kopper
     * @param toBeExecuted The pseuCo file which should be executed by this task.
     */
    constructor(toBeExecuted: File) : this(toBeExecuted, System.out, System.err)

    /**
     * A file representing the `include` folder required by the compiler.
     *
     * @author Konstantin Kopper
     */
    protected val includeDir = File(FileUtilities.getTmpFolder(), "include")

    /**
     * Callback to be called if an exception occurs.
     * Default value just prints the stack trace.
     *
     * @author Konstantin Kopper
     * @since 2.0.0
     */
    protected var onError: (Exception) -> Unit = { e -> e.printStackTrace() }

    /**
     * Assigns [action] to [onError], i.e. [action] is executed whenever an exception occurs.
     *
     * @author Konstantin Kopper
     * @since 2.0.0
     */
    fun onError(action: (Exception) -> Unit) {
        onError = action
    }

    /**
     * Callback to be called if an error was generated by the compiler.
     * Default value just prints the stack trace.
     *
     * @author Konstantin Kopper
     * @since 2.1
     */
    protected var onCompilerError: (Throwable) -> Unit = { e -> e.printStackTrace() }

    /**
     * Assigns [action] to [onCompilerError], i.e. [action] is executed whenever the compiler generates an error.
     *
     * @author Konstantin Kopper
     * @since 2.1
     */
    fun onCompilerError(action: (Throwable) -> Unit) {
        onCompilerError = action
    }

    /**
     * Initialize the pseuCo task.
     * Useful to configure compiler.
     *
     * @author Konstantin Kopper
     * @sicne 2.0.0
     */
    protected open fun initialize() {
        MainCodeGen.generatedJavaCodeFolder = this.includeDir.name
        MainCodeGen.generatedJavaCodeFolderPath = this.includeDir.absolutePath

        // Initialize pseuco-java-compiler, copied from PseuCoCo.java, lines 375 ff.
        MainCodeGen.rootNode = null
        MainCodeGen.output = true
        MainCodeGen.generatedCode = null
        MainCodeGen.isJavaCodeGeneration = true

        CodeGen.mainCodeObject = MainCode()
        CodeGen.monitors = HashMap<String, Code>()
        CodeGen.structs = HashMap<String, Code>()
        CodeGen.usedVariableNames = HashSet()
    }

    /**
     * Unzip the bundled 'include' archive to [includeDir].
     * Can be replaced by custom implementation.
     *
     * @author Konstantin Kopper
     * @since 2.0.0
     */
    protected open fun unzipIncludeDir() = ZipUtilities.unzipResource("/include.zip", includeDir)

    /**
     * Set the standard output streams to the ones given by the constructor.
     *
     * @author Konstantin Kopper
     * @since 2.0.0
     */
    protected fun setOutputStreams() {
        System.setOut(PrintStream(out, true))
        System.setErr(PrintStream(err, true))
    }

    /**
     * Parses the pseuCo program and generates the corresponding Java code.
     *
     * @author Konstantin Kopper
     * @since 2.0.0
     */
    @Throws(PseuCoCompilerError::class, Exception::class)
    protected open fun parsePseuCoProgram() {
        // Call the pseuco parser. This will generate Java code in the 'include' folder.
        val args = arrayOf(MainCodeGen.getPlatformIndependentPath(this.toBeExecuted.absolutePath))

        // Disable assertions for 'tree' package.
        javaClass.classLoader.setPackageAssertionStatus("tree", false)
        javaClass.classLoader.setPackageAssertionStatus("codeGen", false)

        try {
            PseuCoParser.main(args)
        } catch (e: Exception) {
            if (Main.DEBUG && System.err != Main.STANDARD_ERR)
                e.printStackTrace(Main.STANDARD_ERR)

            throw when (e) {
                is TypeIllegalException, is TypeEnvironmentException, is ParseException -> PseuCoCompilerError(e)
                else -> e
            }
        }
    }

    /**
     * Compiles the Java program found in the `include` folder.
     *
     * @author Konstantin Kopper
     * @since 2.0.0
     */
    protected fun compileGeneratedJavaCode(): Boolean {
        val compiler = try {
            ToolProvider.getSystemJavaCompiler()!!
        } catch (e: NullPointerException) {
            System.err.println("Could not compile generated Java code:")
            System.err.println("Maybe you have no JDK installed.")
            if (Main.DEBUG && System.err != Main.STANDARD_ERR)
                Main.STANDARD_ERR.println("No Java compiler provided.")

            // Throw exception to interrupt run method.
            throw e
        }

        val diagnostics = DiagnosticCollector<JavaFileObject>()

        val fileManager = compiler.getStandardFileManager(diagnostics, null, null)

        val files = listOf(*this.includeDir.listFiles()!!)

        val task = compiler.getTask(PrintWriter(Main.STANDARD_ERR), fileManager, diagnostics,
            null, null, fileManager.getJavaFileObjectsFromFiles(files))

        return task.call().also {
            if (Main.DEBUG)
                Main.STANDARD_OUT.println("${if (it) "Successfully compiled" else "Could not compile"} generated Java code.")
            if (!it) {
                if (Main.DEBUG && System.err != Main.STANDARD_ERR) {
                    diagnostics.diagnostics.forEach((Main.STANDARD_ERR)::println)
                }

                throw PseuCoCompilerError(diagnostics.diagnostics.joinToString(separator = "\n") { d -> d.toString() })
            }
        }
    }

    /**
     * Executes the generated and compiled Java code.
     * To be implemented for every specific task.
     *
     * @author Konstantin Kopper
     */
    protected abstract fun executeGeneratedJavaCode()

    /**
     * Calls join on all active threads belonging to [agents].
     *
     * @author Konstantin Kopper
     */
    private fun joinAgents() {
        arrayOfNulls<Thread>(1).also {
            while (agents.enumerate(it) > 0)
                it[0]?.join()
        }
    }

    /**
     * Resets the output streams to the standard ones.
     * Uses [Main.STANDARD_OUT] and [Main.STANDARD_ERR].
     *
     * @author Konstantin Kopper
     */
    protected fun resetOutputStreams() {
        System.out.flush()
        System.setOut(Main.STANDARD_OUT)
        System.err.flush()
        System.setErr(Main.STANDARD_ERR)
    }

    /**
     * Deletes [includeDir].
     *
     * @author Konstantin Kopper
     */
    protected fun removeUnzipped() = FileUtils.deleteQuietly(this.includeDir)

    /**
     * Actual executed method of this task.
     *
     * Calls the helper methods in the following order:
     *  * [initialize]
     *  * [unzipIncludeDir]
     *  * [setOutputStreams]
     *  * [parsePseuCoProgram]
     *  * [compileGeneratedJavaCode]
     *  * [executeGeneratedJavaCode]
     *
     * Even when an error occurred in any of the previous methods, the following are executed in the end:
     *  * [joinAgents]
     *  * [resetOutputStreams]
     *  * [removeUnzipped]
     *
     * @author Konstantin Kopper
     * @see Runnable
     * @since 2.0.0
     */
    override fun run() {
        try {
            initialize()
            unzipIncludeDir()
            setOutputStreams()
            parsePseuCoProgram()
            compileGeneratedJavaCode().takeIf { it } ?: return
            executeGeneratedJavaCode()
        } catch (e: PseuCoCompilerError) {
            onCompilerError(e.cause ?: e)
        } catch (e: Exception) {
            onError(e)
        } finally {
            joinAgents()
            resetOutputStreams()
            removeUnzipped()
        }
    }

    override fun interrupt() = interruptAgents()

    internal companion object {
        /**
         * Thread group collecting all pseuCo agents.
         *
         * @author Konstantin Kopper
         * @since 2.0.0
         * @see ThreadGroup
         */
        internal val agents = ThreadGroup("pseuCo Agents")

        /**
         * Calls interrupt on all [agents].
         *
         * @author Konstantin Kopper
         * @since 2.0.0
         * @see Thread.interrupt
         */
        internal fun interruptAgents() = agents.interrupt()
    }
}
