package jani.interaction.basic

import kotlinx.serialization.decodeFromString
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.Json
import kotlinx.serialization.json.JsonArray
import kotlinx.serialization.json.JsonPrimitive
import kotlinx.serialization.json.decodeFromJsonElement
import kotlinx.serialization.json.encodeToJsonElement
import kotlinx.serialization.json.jsonObject
import kotlinx.serialization.json.jsonPrimitive
import org.junit.jupiter.api.Assertions.assertEquals
import org.junit.jupiter.api.Assertions.assertIterableEquals
import org.junit.jupiter.api.Assertions.assertTrue
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertAll
import org.junit.jupiter.api.assertDoesNotThrow

@DisplayName("jani.interaction.basic.ParameterType serialization")
internal class ParameterTypeSerializationTest {

    @Test
    @DisplayName("Serialize 'Bool' 1")
    fun testSerializeBool1() {
        val pt: ParameterType = ParameterType.Bool
        val msg = Json.encodeToString(ParameterType.Companion, pt)

        assertEquals("\"bool\"", msg)
    }

    @Test
    @DisplayName("Serialize 'Bool' 2")
    fun testSerializeBool2() {
        val pt = ParameterType.Bool
        val msg = Json.encodeToString<ParameterType>(pt)

        assertEquals("\"bool\"", msg)
    }

    @Test
    @DisplayName("Deserialize 'Bool'")
    fun testDeserializeBool() {
        val msg = JsonPrimitive("bool")
        val parameterType = Json.decodeFromJsonElement<ParameterType>(msg)

        assertTrue(parameterType is ParameterType.Bool)
    }

    @Test
    @DisplayName("Serialize 'Int' 1")
    fun testSerializeInt1() {
        val parameterType: ParameterType = ParameterType.Int
        val msg = Json.encodeToString(parameterType)

        assertEquals("\"int\"", msg)
    }

    @Test
    @DisplayName("Serialize 'Int' 2")
    fun testSerializeInt2() {
        val parameterType = ParameterType.Int
        val msg = Json.encodeToString<ParameterType>(parameterType)

        assertEquals("\"int\"", msg)
    }

    @Test
    @DisplayName("Deserialize 'Int'")
    fun testDeserializeInt() {
        val msg = JsonPrimitive("int")
        val parameterType = Json.decodeFromJsonElement<ParameterType>(msg)

        assertTrue(parameterType is ParameterType.Int)
    }

    @Test
    @DisplayName("Serialize 'Real' 1")
    fun testSerializeReal1() {
        val parameterType: ParameterType = ParameterType.Real
        val msg = Json.encodeToString(parameterType)

        assertEquals("\"real\"", msg)
    }

    @Test
    @DisplayName("Serialize 'Real' 2")
    fun testSerializeReal2() {
        val parameterType = ParameterType.Real
        val msg = Json.encodeToString<ParameterType>(parameterType)

        assertEquals("\"real\"", msg)
    }

    @Test
    @DisplayName("Deserialize 'Real'")
    fun testDeserializeReal() {
        val msg = JsonPrimitive("real")
        val parameterType = Json.decodeFromJsonElement<ParameterType>(msg)

        assertTrue(parameterType is ParameterType.Real)
    }

    @Test
    @DisplayName("Serialize 'String' 1")
    fun testSerializeString1() {
        val parameterType: ParameterType = ParameterType.String
        val msg = Json.encodeToString(parameterType)

        assertEquals("\"string\"", msg)
    }

    @Test
    @DisplayName("Serialize 'String' 2")
    fun testSerializeString2() {
        val parameterType = ParameterType.String
        val msg = Json.encodeToString<ParameterType>(parameterType)

        assertEquals("\"string\"", msg)
    }

    @Test
    @DisplayName("Deserialize 'String'")
    fun testDeserializeString() {
        val msg = JsonPrimitive("string")
        val parameterType = Json.decodeFromJsonElement<ParameterType>(msg)

        assertTrue(parameterType is ParameterType.String)
    }

    @Test
    @DisplayName("Serialize 'Enum' 1")
    fun testSerializeEnum1() {
        val parameterType: ParameterType = ParameterType.Enum("bool", "int")
        val msg = Json.encodeToString(parameterType)

        assertEquals("[\"bool\",\"int\"]", msg)
    }

    @Test
    @DisplayName("Serialize 'Enum' 2")
    fun testSerializeEnum2() {
        val parameterType = ParameterType.Enum("bool", "int")
        val msg = Json.encodeToString<ParameterType>(parameterType)

        assertEquals("[\"bool\",\"int\"]", msg)
    }

    @Test
    @DisplayName("Deserialize 'Enum'")
    fun testDeserializeEnum() {
        val msg = JsonArray(listOf(JsonPrimitive("bool"), JsonPrimitive("int")))
        val parameterType = Json.decodeFromJsonElement<ParameterType>(msg)

        assertTrue(parameterType is ParameterType.Enum)
        assertIterableEquals(listOf("bool", "int"), parameterType as ParameterType.Enum)
    }

    @Test
    @DisplayName("Serialize 'Option' 1")
    fun testSerializeOption1() {
        val parameterType: ParameterType = ParameterType.Option(ParameterType.Real)
        val msg = Json.encodeToJsonElement(parameterType)

        val o = assertDoesNotThrow { msg.jsonObject }

        assertAll(
            {
                assertTrue(o.containsKey("kind"))
                assertEquals("option", o["kind"]!!.jsonPrimitive.content)
            },
            {
                assertTrue(o.containsKey("type"))
                assertEquals("real", o["type"]!!.jsonPrimitive.content)
            }
        )
    }

    @Test
    @DisplayName("Serialize 'Option' 2")
    fun testSerializeOption2() {
        val parameterType = ParameterType.Option(ParameterType.Real)
        val msg = Json.encodeToJsonElement<ParameterType>(parameterType)

        val o = assertDoesNotThrow { msg.jsonObject }

        assertAll(
            {
                assertTrue(o.containsKey("kind"))
                assertEquals("option", o["kind"]!!.jsonPrimitive.content)
            },
            {
                assertTrue(o.containsKey("type"))
                assertEquals("real", o["type"]!!.jsonPrimitive.content)
            }
        )
    }

    @Test
    @DisplayName("Deserialize 'Option'")
    fun testDeserializeOption() {
        val msg = """{"kind":"option","type":"real"}"""
        val parameterType = Json.decodeFromString<ParameterType>(msg)

        assertTrue(parameterType is ParameterType.Option)
        assertEquals(ParameterType.Real, (parameterType as ParameterType.Option).type)
    }
}
